@@ -42,7 +42,7 @@ pub fn batch_to_concat(circuit: CircuitRc, axis: usize, num_batches: usize) -> C
                 sub.clone(),
                 ModuleNodeArgSpec {
                     symbol: set_named_axes(
-                        &Symbol::new_with_random_uuid(sub.info().shape.clone(), None),
+                        Symbol::new_with_random_uuid(sub.info().shape.clone(), None),
                         BTreeMap::from([(*i as u8, name.clone())]),
                     ),
                     batchable: true,
@@ -174,7 +174,7 @@ pub fn batch_einsum(einsum: &Einsum, context: &mut OptimizationContext) -> CircR
         .map(|(i, (c, ints))| {
             (
                 set_named_axes(
-                    &Symbol::new_with_random_uuid(c.info().shape.clone(), Some(i.to_string())),
+                    Symbol::new_with_random_uuid(c.info().shape.clone(), Some(i.to_string())),
                     ints.iter()
                         .enumerate()
                         .map(|(i, int)| (i as u8, int.to_string()))
