@@ -21,11 +21,11 @@ pub fn numel_sort_key(node: &CircuitRc) -> Vec<u8> {
 #[pyfunction]
 #[pyo3(name = "canonicalize_node")]
 pub fn canonicalize_node_py(circuit: CircuitRc) -> CircuitRc {
-    canonicalize_node_op(&circuit).unwrap_or(circuit.clone())
+    canonicalize_node_op(circuit.clone()).unwrap_or(circuit)
 }
 
-pub fn canonicalize_node_op(circuit: &Circuit) -> Option<CircuitRc> {
-    match &*circuit {
+pub fn canonicalize_node_op(circuit: CircuitRc) -> Option<CircuitRc> {
+    match &**circuit {
         Circuit::Rearrange(rearrange) => Some(Rearrange::nrc(
             rearrange.node.clone(),
             rearrange
@@ -67,10 +67,10 @@ pub fn deep_canonicalize_py(circuit: CircuitRc) -> CircuitRc {
 
 pub fn deep_canonicalize(circuit: CircuitRc, context: &mut OptimizationContext) -> CircuitRc {
     deep_map_op_context(
-        &circuit,
+        circuit.clone(),
         &|x, _c: &mut HashMap<(), ()>| canonicalize_node_op(x),
         &mut HashMap::<(), ()>::new(),
         &mut context.cache.canonicalized,
     )
-    .unwrap_or(circuit.clone())
+    .unwrap_or(circuit)
 }
