@@ -7,10 +7,10 @@ use super::{visit_circuit_fallable, CircuitNode, CircuitNodeUnion, CircuitRc};
 use crate::hashmaps::AHashSet as HashSet;
 use crate::pyo3_prelude::*;
 
-type RawMatcher = Arc<dyn Fn(CircuitRc) -> Result<bool, PyErr>>;
+type RawMatcher = Arc<dyn Fn(CircuitRc) -> Result<bool, PyErr> + Send + Sync>;
 
 #[derive(Clone)]
-#[pyclass(unsendable)]
+#[pyclass]
 pub struct Matcher(RawMatcher);
 #[pymethods]
 impl Matcher {
