@@ -1,7 +1,8 @@
 #![allow(clippy::borrow_deref_ref)]
+use crate::all_imports::{RInnerInts, Sv, TensorInvariantError};
 use crate::circuit::circuit_utils::toposort_circuit;
 use crate::hashmaps::{AHashSet as HashSet, FxHashMap as HashMap};
-use crate::pyo3_prelude::*;
+use crate::util::AxisInt;
 use crate::{
     cached_lambda,
     py_types::{make_bytes_py, PyShape, Tensor},
@@ -11,6 +12,7 @@ use crate::{
     },
     util::AsOp,
 };
+use crate::{pyo3_prelude::*, sv};
 pub use computational_node::{
     flat_concat, flat_concat_back, Add, Concat, Einsum, EinsumAxes, Index, Rearrange, Scatter,
 };
@@ -31,7 +33,6 @@ use std::sync::Arc;
 use std::{
     iter::zip,
     ops::{Deref, DerefMut},
-    rc::Rc,
 };
 use uuid::uuid;
 pub use variable_nodes::{AutoTag, DiscreteVar, StoredCumulantVar};
@@ -73,20 +74,97 @@ pub mod scheduling_z3;
 pub mod variable_nodes;
 
 mod circuit_node_private {
+    use super::{CachedCircuitInfo, CircuitConstructionError};
     pub trait CircuitNodePrivate {
-        fn info_mut(&mut self) -> &mut super::CachedCircuitInfo;
+        fn info_mut(&mut self) -> &mut CachedCircuitInfo;
         fn name_mut(&mut self) -> &mut Option<String>;
     }
+
+    pub trait CircuitNodeInit {
+        fn init_info_impl(self) -> Result<Self, CircuitConstructionError>
+        where
+            Self: Sized;
+
+        fn rename_impl(self, new_name: Option<String>) -> Self
+        where
+            Self: Sized;
+
+        fn update_info_impl<F>(self, f: F) -> Result<Self, CircuitConstructionError>
+        where
+            Self: Sized,
+            F: FnOnce(&mut CachedCircuitInfo);
+    }
 }
 use circuit_node_private::*;
 
+impl<T: CircuitNodePrivate + CircuitNode> CircuitNodeInit for T {
+    fn init_info_impl(mut self) -> Result<Self, CircuitConstructionError>
+    where
+        Self: Sized,
+    {
+        self.info_mut().shape = self.compute_shape(); // set shape so methods to compute other info can use it
+
+        let mut hasher = self.compute_hash();
+        self.info_mut().named_axes = self.compute_named_axes();
+        hasher.update(self.name().unwrap_or("").as_bytes());
+        hasher.update(&self.node_type_uuid());
+        hasher.update(uuid!("025e9af4-1366-4211-aa5f-7c28fc6cdf9f").as_bytes());
+        for (axis, name) in &self.info().named_axes {
+            if *axis as usize >= self.info().shape.len() {
+                return Err(CircuitConstructionError::NamedAxisAboveRank {});
+            }
+            hasher.update(&[*axis]);
+            hasher.update(name.as_bytes());
+        }
+        self.info_mut().is_constant = self.compute_is_constant();
+        self.info_mut().is_explicitly_computable = self.compute_is_explicitly_computable();
+        self.info_mut().can_be_sampled = self.compute_can_be_sampled();
+        self.info_mut().hash = hasher.finalize().into();
+        self.info_mut().max_non_input_size = self.max_non_input_size();
+        self.info_mut().device_dtype = self.compute_device_dtype()?;
+        Ok(self)
+    }
+
+    fn rename_impl(mut self, new_name: Option<String>) -> Self
+    where
+        Self: Sized,
+    {
+        *self.name_mut() = new_name;
+        self.init_info_impl().unwrap() // we could avoid recomputing some stuff if we wanted
+    }
+
+    fn update_info_impl<F>(mut self, f: F) -> Result<Self, CircuitConstructionError>
+    where
+        Self: Sized,
+        F: FnOnce(&mut CachedCircuitInfo),
+    {
+        f(self.info_mut());
+        self.init_info_impl()
+    }
+}
+
 use self::{
     circuit_utils::total_flops,
     print::{print_circuit_stats, repr_circuit_deep_compiler},
 };
 
-pub type NamedAxes = BTreeMap<u8, String>;
-pub trait CircuitNode: CircuitNodePrivate {
+pub type Idxs = Sv<[usize; 6]>;
+
+fn check_canon_idxs(count: usize, ints: &[i64]) -> Result<Vec<usize>, i64> {
+    let icount = count as i64;
+    ints.iter()
+        .map(|&i| {
+            if i >= icount || i < -icount {
+                Err(i)
+            } else {
+                Ok(((i + icount) % icount) as usize)
+            }
+        })
+        .collect()
+}
+
+pub type NamedAxes = BTreeMap<AxisInt, String>;
+pub trait CircuitNode: CircuitNodeInit {
     // ==== implementable section ===
     //
     // NOTE: ALL FNS IN THIS SECTION *MUST* BE COPIED TO THE CIRCUIT NODE UNION IMPL!
@@ -142,6 +220,28 @@ pub trait CircuitNode: CircuitNodePrivate {
         self.info().numel_usize()
     }
 
+    fn c(self) -> Circuit;
+    fn rc(self) -> CircuitRc;
+    fn init_info(self) -> Result<Self, CircuitConstructionError>
+    where
+        Self: Sized,
+    {
+        self.init_info_impl()
+    }
+    fn rename(self, new_name: Option<String>) -> Self
+    where
+        Self: Sized,
+    {
+        self.rename_impl(new_name)
+    }
+    fn update_info<F>(self, f: F) -> Result<Self, CircuitConstructionError>
+    where
+        Self: Sized,
+        F: FnOnce(&mut CachedCircuitInfo),
+    {
+        self.update_info_impl(f)
+    }
+
     // ==== default section ===
     // FUNCTIONS BELOW HERE *shouldn't* be overridden by implementors!
     // (if you do implement, this won't be picked up on by union types!)
@@ -245,41 +345,6 @@ pub trait CircuitNode: CircuitNodePrivate {
         }
     }
 
-    fn init_info(mut self) -> Result<Self, CircuitConstructionError>
-    where
-        Self: Sized,
-    {
-        self.info_mut().shape = self.compute_shape(); // set shape so methods to compute other info can use it
-
-        let mut hasher = self.compute_hash();
-        self.info_mut().named_axes = self.compute_named_axes();
-        hasher.update(self.name().unwrap_or("").as_bytes());
-        hasher.update(&self.node_type_uuid());
-        hasher.update(uuid!("025e9af4-1366-4211-aa5f-7c28fc6cdf9f").as_bytes());
-        for (axis, name) in &self.info().named_axes {
-            if *axis as usize >= self.info().shape.len() {
-                return Err(CircuitConstructionError::NamedAxisAboveRank {});
-            }
-            hasher.update(&[*axis]);
-            hasher.update(name.as_bytes());
-        }
-        self.info_mut().is_constant = self.compute_is_constant();
-        self.info_mut().is_explicitly_computable = self.compute_is_explicitly_computable();
-        self.info_mut().can_be_sampled = self.compute_can_be_sampled();
-        self.info_mut().hash = hasher.finalize().into();
-        self.info_mut().max_non_input_size = self.max_non_input_size();
-        self.info_mut().device_dtype = self.compute_device_dtype()?;
-        Ok(self)
-    }
-
-    fn rename(mut self, new_name: Option<String>) -> Self
-    where
-        Self: Sized,
-    {
-        *self.name_mut() = new_name;
-        self.init_info().unwrap() // we could avoid recomputing some stuff if we wanted
-    }
-
     fn max_non_input_size(&self) -> BigUint {
         self.children()
             .map(|x| x.info().max_non_input_size.clone())
@@ -288,30 +353,16 @@ pub trait CircuitNode: CircuitNodePrivate {
             .unwrap_or(0usize.into())
     }
 
-    fn c(self) -> Circuit
-    where
-        Self: Into<Circuit>,
-    {
-        self.into()
-    }
-
-    fn rc(self) -> CircuitRc
-    where
-        Self: Into<Circuit>,
-    {
-        CircuitRc(Arc::new(self.c()))
-    }
-
     fn compiler_repr(&self) -> String
     where
-        Self: Into<Circuit> + Clone,
+        Self: Clone,
     {
         repr_circuit_deep_compiler(&self.clone().c(), false, false)
     }
 
     fn compiler_print(&self)
     where
-        Self: Into<Circuit> + Clone,
+        Self: Clone,
     {
         println!("{}", self.compiler_repr())
     }
@@ -319,22 +370,149 @@ pub trait CircuitNode: CircuitNodePrivate {
     fn get_hash(&self) -> HashBytes {
         self.info().hash
     }
+
+    fn check_canon_axes(&self, axes: &[i64]) -> Result<RInnerInts, CircuitConstructionError> {
+        check_canon_idxs(self.info().rank(), axes)
+            .map(|v| v.into_iter().map(|x| x as u8).collect())
+            .map_err(|axis| CircuitConstructionError::ReductionAxisOutOfBounds {
+                axis,
+                node_rank: self.info().rank(),
+            })
+    }
+
+    fn sum(&self, axes: &[i64], name: Option<String>) -> Result<Einsum, CircuitConstructionError>
+    where
+        Self: Clone,
+    {
+        let axes = self.check_canon_axes(axes)?;
+        Ok(Einsum::try_new(
+            vec![(self.clone().rc(), (0u8..self.info().rank() as u8).collect())],
+            (0u8..self.info().rank() as u8)
+                .filter(|i| !axes.contains(&i))
+                .collect(),
+            name,
+        )
+        .unwrap())
+    }
+
+    fn mean(&self, axes: &[i64], name: Option<String>) -> Result<Einsum, CircuitConstructionError>
+    where
+        Self: Clone,
+    {
+        let total_size: usize = self
+            .check_canon_axes(axes)?
+            .into_iter()
+            .map(|i| self.info().shape[i as usize])
+            .product();
+        self.sum(axes, name)?.mul(
+            ScalarConstant::nrc(1. / (total_size as f64), sv![], None),
+            None,
+        )
+    }
+
+    fn reduce(
+        &self,
+        op_name: String,
+        axes: &[i64],
+        name: Option<String>,
+    ) -> Result<Circuit, CircuitConstructionError>
+    where
+        Self: Clone,
+    {
+        match op_name.as_str() {
+            "mean" => return self.mean(axes, name).map(CircuitNode::c),
+            "sum" => return self.sum(axes, name).map(CircuitNode::c),
+            _ => (),
+        }
+
+        let axes = self.check_canon_axes(axes)?;
+
+        Ok(GeneralFunction::new_by_name(
+            vec![Rearrange::nrc(
+                self.clone().rc(),
+                RearrangeSpec::combine_axes_at_end(self.info().rank() as u8, axes),
+                None,
+            )],
+            op_name,
+            name,
+        )
+        .unwrap()
+        .c())
+    }
+    fn min_(&self, axes: &[i64], name: Option<String>) -> Result<Circuit, CircuitConstructionError>
+    where
+        Self: Clone,
+    {
+        self.reduce("min".to_owned(), axes, name)
+    }
+    fn max_(&self, axes: &[i64], name: Option<String>) -> Result<Circuit, CircuitConstructionError>
+    where
+        Self: Clone,
+    {
+        self.reduce("max".to_owned(), axes, name)
+    }
+    fn add(&self, other: CircuitRc, name: Option<String>) -> Result<Add, CircuitConstructionError>
+    where
+        Self: Clone,
+    {
+        Add::try_new(vec![self.clone().rc(), other], name)
+    }
+    fn sub(&self, other: CircuitRc, name: Option<String>) -> Result<Add, CircuitConstructionError>
+    where
+        Self: Clone,
+    {
+        self.add(Einsum::scalar_mul(other, -1.0, None).rc(), name)
+    }
+    fn mul(
+        &self,
+        other: CircuitRc,
+        name: Option<String>,
+    ) -> Result<Einsum, CircuitConstructionError>
+    where
+        Self: Clone,
+    {
+        Einsum::elementwise_broadcasted(vec![self.clone().rc(), other], name)
+    }
+    fn mul_scalar(
+        &self,
+        scalar: f64,
+        name: Option<String>,
+        scalar_name: Option<String>,
+    ) -> Result<Einsum, CircuitConstructionError>
+    where
+        Self: Clone,
+    {
+        self.mul(ScalarConstant::nrc(scalar, sv![], scalar_name), name)
+    }
+    fn index(
+        &self,
+        index: TensorIndex,
+        name: Option<String>,
+    ) -> Result<Index, CircuitConstructionError>
+    where
+        Self: Clone,
+    {
+        Index::try_new(self.clone().rc(), index, name)
+    }
 }
 
 pub trait CircuitNodeAutoName: CircuitNode {
     fn auto_name(&self, name: Option<String>) -> Option<String>;
 }
 
-pub trait CircuitNodeUnion {
+pub trait CircuitNodeDefer: CircuitNodeInit {
     fn as_trait_obj(&self) -> &dyn CircuitNode;
-    fn as_trait_obj_mut(&mut self) -> &mut dyn CircuitNode;
     fn map_children_enumerate_impl<F, E>(&self, f: F) -> Result<Self, CircuitConstructionError>
     where
         Self: Sized,
         CircuitConstructionError: From<E>,
         F: FnMut(usize, CircuitRc) -> Result<CircuitRc, E>;
+    fn custom_c(self) -> Circuit;
+    fn custom_rc(self) -> CircuitRc;
+}
+
+pub trait CircuitNodeUnion {
     fn variant_string(&self) -> String;
-    fn node_type_uuid(&self) -> [u8; 16];
 }
 
 // not really needed to be so pedantic with ::std::...
@@ -441,18 +619,38 @@ macro_rules! define_circuit_union_impl {
             )*
         }
 
-        impl $crate::circuit::CircuitNodeUnion for $name {
-            #[inline] // hopefully inlined away?
-            fn as_trait_obj(&self) -> &dyn $crate::circuit::CircuitNode {
+        impl $crate::circuit::CircuitNodeInit for $name {
+            fn init_info_impl(self) -> Result<Self, $crate::circuit::CircuitConstructionError> {
                 match self {
                     $(
-                        Self::$x(node) => node,
+                        Self::$x(node) => Ok(Self::$x(node.init_info()?)),
                     )*
                 }
             }
 
+            fn rename_impl(self, new_name: Option<String>) -> Self {
+                match self {
+                    $(
+                        Self::$x(node) => Self::$x(node.rename(new_name)),
+                    )*
+                }
+            }
+
+            fn update_info_impl<F>(self, f: F) -> Result<Self, $crate::circuit::CircuitConstructionError>
+            where
+                F: FnOnce(&mut $crate::circuit::CachedCircuitInfo),
+            {
+                match self {
+                    $(
+                        Self::$x(node) => Ok(Self::$x(node.update_info(f)?)),
+                    )*
+                }
+            }
+        }
+
+        impl $crate::circuit::CircuitNodeDefer for $name {
             #[inline] // hopefully inlined away?
-            fn as_trait_obj_mut(&mut self) -> &mut dyn $crate::circuit::CircuitNode {
+            fn as_trait_obj(&self) -> &dyn $crate::circuit::CircuitNode {
                 match self {
                     $(
                         Self::$x(node) => node,
@@ -472,6 +670,24 @@ macro_rules! define_circuit_union_impl {
                 }
             }
 
+            fn custom_c(self) -> $crate::circuit::Circuit {
+                match self {
+                    $(
+                        Self::$x(node) => node.c(),
+                    )*
+                }
+            }
+
+            fn custom_rc(self) -> $crate::circuit::CircuitRc {
+                match self {
+                    $(
+                        Self::$x(node) => node.rc(),
+                    )*
+                }
+            }
+        }
+
+        impl $crate::circuit::CircuitNodeUnion for $name {
             fn variant_string(&self) -> String {
                 match self {
                     $(
@@ -479,13 +695,6 @@ macro_rules! define_circuit_union_impl {
                     )*
                 }
             }
-            fn node_type_uuid(&self) ->[u8;16]{
-                match self {
-                    $(
-                        Self::$x(node) => node.node_type_uuid(),
-                    )*
-                }
-            }
 
         }
 
@@ -657,20 +866,17 @@ macro_rules! circuit_node_auto_impl {
         fn node_type_uuid(&self) -> [u8; 16] {
             *uuid::uuid!($the_uuid).as_bytes()
         }
+        fn c(self) -> $crate::circuit::Circuit {
+            self.into()
+        }
+        fn rc(self) -> $crate::circuit::CircuitRc {
+            $crate::circuit::CircuitRc(std::sync::Arc::new(self.c()))
+        }
     };
 }
 
-impl<T: CircuitNodeUnion> CircuitNodePrivate for T {
-    fn info_mut(&mut self) -> &mut CachedCircuitInfo {
-        self.as_trait_obj_mut().info_mut()
-    }
-    fn name_mut(&mut self) -> &mut Option<String> {
-        self.as_trait_obj_mut().name_mut()
-    }
-}
-
 // UPDATE ME WHEN YOU CHANGE CircuitNode Trait!!!
-impl<T: CircuitNodeUnion> CircuitNode for T {
+impl<T: CircuitNodeDefer> CircuitNode for T {
     fn info(&self) -> &CachedCircuitInfo {
         self.as_trait_obj().info()
     }
@@ -739,6 +945,14 @@ impl<T: CircuitNodeUnion> CircuitNode for T {
     fn intermediate_cost_bound(&self) -> usize {
         self.as_trait_obj().intermediate_cost_bound()
     }
+
+    fn c(self) -> Circuit {
+        self.custom_c()
+    }
+
+    fn rc(self) -> CircuitRc {
+        self.custom_rc()
+    }
 }
 
 pub type HashBytes = [u8; 32];
@@ -808,6 +1022,8 @@ pub enum CircuitConstructionError {
 
     #[error("StoredCumulantVar invalid cumulant number {number}")]
     StoredCumulantVarInvalidCumulantNumber { number: usize },
+    #[error("TensorInvariantError {err:?}")]
+    TensorInvariantError { err: TensorInvariantError },
 
     #[error("StoredCumulantVar cumulant number {cumulant_number} needs to be base shape, {base_shape:?} times cumulant number, got {cumulant_shape:?}")]
     StoredCumulantVarCumulantWrongShape {
@@ -863,6 +1079,9 @@ pub enum CircuitConstructionError {
     #[error("index rank too high: {index_rank} vs {node_rank}")]
     IndexRankTooHigh { index_rank: usize, node_rank: usize },
 
+    #[error("reduction axis out of bounds: {axis} vs {node_rank}")]
+    ReductionAxisOutOfBounds { axis: i64, node_rank: usize },
+
     #[error("Index {axis:?} out of bounds, index: {index:?} shape: {shape:?}. NB: Rust circuit slices don't saturate like Python ones do.")]
     IndexOutOfBounds {
         index: TensorIndex,
@@ -1467,6 +1686,47 @@ impl DerefMut for CircuitRc {
     }
 }
 
+impl CircuitNodeInit for CircuitRc {
+    fn init_info_impl(self) -> Result<Self, CircuitConstructionError> {
+        Ok(self.c().clone().init_info()?.rc())
+    }
+
+    fn rename_impl(self, new_name: Option<String>) -> Self {
+        self.c().clone().rename(new_name).rc()
+    }
+
+    fn update_info_impl<F>(self, f: F) -> Result<Self, CircuitConstructionError>
+    where
+        F: FnOnce(&mut CachedCircuitInfo),
+    {
+        Ok(self.c().clone().update_info_impl(f)?.rc())
+    }
+}
+
+impl CircuitNodeDefer for CircuitRc {
+    fn as_trait_obj(&self) -> &dyn CircuitNode {
+        // deref to avoid infinite recursion
+        (**self).as_trait_obj()
+    }
+    fn map_children_enumerate_impl<F, E>(&self, f: F) -> Result<Self, CircuitConstructionError>
+    where
+        Self: Sized,
+        CircuitConstructionError: From<E>,
+        F: FnMut(usize, CircuitRc) -> Result<CircuitRc, E>,
+    {
+        (**self).map_children_enumerate(f).map(CircuitNode::rc)
+    }
+
+    fn custom_c(self) -> Circuit {
+        (**self).clone()
+    }
+
+    // fast custom impl
+    fn custom_rc(self) -> CircuitRc {
+        self
+    }
+}
+
 #[pyclass(subclass, name = "Circuit")]
 #[derive(Clone, Debug)]
 pub struct PyCircuitBase(Arc<Circuit>);
@@ -1508,6 +1768,20 @@ fn python_apply_fn_to_sub<N: CircuitNode>(
     })
 }
 
+#[derive(Clone, Debug, FromPyObject)]
+enum PyReductionAxes {
+    Single(i64),
+    Many(Vec<i64>),
+}
+
+fn reduction_to_ints(x: Option<PyReductionAxes>, ndim: usize) -> Vec<i64> {
+    match x {
+        None => (0..ndim as i64).collect(),
+        Some(PyReductionAxes::Single(x)) => vec![x],
+        Some(PyReductionAxes::Many(x)) => x,
+    }
+}
+
 #[pymethods]
 impl PyCircuitBase {
     #[getter]
@@ -1588,12 +1862,12 @@ impl PyCircuitBase {
         Ok(out)
     }
 
-    #[args(bijection = "false", shape_only_necessary = "false")]
+    #[args(bijection = "true", shape_only_necessary = "false")]
     fn compiler_repr(&self, bijection: bool, shape_only_necessary: bool) -> String {
         repr_circuit_deep_compiler(self, bijection, shape_only_necessary)
     }
 
-    #[args(bijection = "false", shape_only_necessary = "false")]
+    #[args(bijection = "true", shape_only_necessary = "false")]
     fn compiler_print(&self, bijection: bool, shape_only_necessary: bool) {
         println!("{}", self.compiler_repr(bijection, shape_only_necessary))
     }
@@ -1619,6 +1893,87 @@ impl PyCircuitBase {
     fn evaluate(&self, device_dtype: TorchDeviceDtypeOp) -> Result<Tensor, TensorEvalError> {
         evaluate_fn_dtype_device((***self).clone().rc(), device_dtype)
     }
+
+    fn reduce(
+        &self,
+        op_name: String,
+        axis: Option<PyReductionAxes>,
+        name: Option<String>,
+    ) -> Result<Circuit, CircuitConstructionError> {
+        self.0
+            .reduce(op_name, &reduction_to_ints(axis, self.info().rank()), name)
+    }
+
+    // sadly, below can't be defined with macro due to outer proc macro...
+    fn sum(
+        &self,
+        axis: Option<PyReductionAxes>,
+        name: Option<String>,
+    ) -> Result<CircuitRc, CircuitConstructionError> {
+        Ok(self
+            .0
+            .sum(&reduction_to_ints(axis, self.info().rank()), name)?
+            .rc())
+    }
+    fn mean(
+        &self,
+        axis: Option<PyReductionAxes>,
+        name: Option<String>,
+    ) -> Result<CircuitRc, CircuitConstructionError> {
+        Ok(self
+            .0
+            .mean(&reduction_to_ints(axis, self.info().rank()), name)?
+            .rc())
+    }
+    fn min(
+        &self,
+        axis: Option<PyReductionAxes>,
+        name: Option<String>,
+    ) -> Result<CircuitRc, CircuitConstructionError> {
+        Ok(self
+            .0
+            .min_(&reduction_to_ints(axis, self.info().rank()), name)?
+            .rc())
+    }
+    fn max(
+        &self,
+        axis: Option<PyReductionAxes>,
+        name: Option<String>,
+    ) -> Result<CircuitRc, CircuitConstructionError> {
+        Ok(self
+            .0
+            .max_(&reduction_to_ints(axis, self.info().rank()), name)?
+            .rc())
+    }
+
+    fn add(&self, other: CircuitRc, name: Option<String>) -> Result<Add, CircuitConstructionError> {
+        self.0.add(other, name)
+    }
+    fn sub(&self, other: CircuitRc, name: Option<String>) -> Result<Add, CircuitConstructionError> {
+        self.0.sub(other, name)
+    }
+    fn mul(
+        &self,
+        other: CircuitRc,
+        name: Option<String>,
+    ) -> Result<Einsum, CircuitConstructionError> {
+        self.0.mul(other, name)
+    }
+    fn mul_scalar(
+        &self,
+        scalar: f64,
+        name: Option<String>,
+        scalar_name: Option<String>,
+    ) -> Result<Einsum, CircuitConstructionError> {
+        self.0.mul_scalar(scalar, name, scalar_name)
+    }
+    fn index(
+        &self,
+        index: TensorIndex,
+        name: Option<String>,
+    ) -> Result<Index, CircuitConstructionError> {
+        self.0.index(index, name)
+    }
 }
 
 // TODO: prelude!
