@@ -1,7 +1,6 @@
 use rust_circuit::{
     circuit::{
         circuit_optimizer::{optimize_circuit, OptimizationContext, OptimizationSettings},
-        deep_rewrite::compiler_simp,
         parse_compiler_repr_bijection,
     },
     rrfs::get_rrfs_dir,
@@ -9,6 +8,7 @@ use rust_circuit::{
 use std::fs;
 
 #[test]
+#[ignore]
 fn benchmarking_circs() {
     fn item() {
         pyo3::prepare_freethreaded_python();
@@ -34,7 +34,7 @@ fn benchmarking_circs() {
         settings.log_simplifications = true;
         let mut context = OptimizationContext::new_settings(settings);
         for circuit in circuits {
-            let result = optimize_circuit(circuit, &mut context);
+            let _result = optimize_circuit(circuit, &mut context);
             println!("{}", context.stringify_logs());
             // println!("{:?}", result.info().hash);
         }
